import argparse


class ArgumentTypeBase:
    class _NA:
        pass

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self._default = self._NA
        self.__name__ = self.__class__.__name__

    @property
    def default(self):
        return self._default

    @default.setter
    def default(self, value):
        if value is ArgumentTypeBase._NA:
            self._default = ArgumentTypeBase._NA
        else:
            self._default = self._validate(value)
        return self._default

    def __call__(self, txt):
        self._value = None
        self._value = self._validate(txt)
        return self

    @property
    def value(self):
        return getattr(self, "_value", self.default)

    def _validate(self, value):
        try:
            return self.validate(value)
        except argparse.ArgumentTypeError as exc:
            if not hasattr(self, "_value"):
                raise RuntimeError(f"cannot use {value=} as default: {exc.args[0]}")
            raise

    def validate(self, txt):
        raise NotImplementedError("need to implement the .validate(self, txt)  method")



from __future__ import annotations

from collections.abc import Iterable, MutableMapping
from typing import Any

DELETE = object()


def dictupdate(data: MutableMapping[str, Any], updates: Iterable[tuple[str, Any]]) -> list[tuple[str, Any]]:
    undo = []
    for path, value in updates:
        cur = data
        loc = path.split(".")
        for i in range(len(loc) - 1):
            if loc[i] not in cur:
                cur[loc[i]] = dict()
                undo.append((".".join(loc[: i + 1]), DELETE))
            cur = cur[loc[i]]

        orig = cur.get(loc[-1], DELETE)
        undo.append((path, orig))
        if value is DELETE:
            del cur[loc[-1]]
        else:
            cur[loc[-1]] = value

    return undo


def dictrollback(data: MutableMapping[str, Any], undo: list[tuple[str, Any]]) -> None:
    for path, value in reversed(undo):
        cur = data
        loc = path.split(".")
        for c in loc[:-1]:
            cur = cur[c]
        if value is DELETE:
            del cur[loc[-1]]
        else:
            cur[loc[-1]] = value



from __future__ import annotations

import dataclasses as dc
import functools
import itertools
from enum import IntEnum, auto


class S(IntEnum):
    OK = auto()
    FAILED = auto()
    WARN = auto()
    NOSTATUS = auto()


@dc.dataclass
class Record:
    status: S
    group: str  # kry to group the record under
    key: str
    report: str = ""


@dc.dataclass
class ColorText:
    message: str
    status: S

    def __format__(self, width):
        reset = "\033[0m"
        message = self.message
        msg = {
            S.OK: f"\033[42m{message}{reset}",
            S.FAILED: f"\033[41m{message}{reset}",
            S.WARN: f"\033[43;30m{message}{reset}",
            S.NOSTATUS: f"{message}",
        }[self.status]
        return msg + " " * (int(width or len(self.message)) - len(self.message))


def indent(txt: str, pre: str, first: str | None = None) -> str:
    return (pre if first is None else first) + txt.replace("\n", "\n" + pre)


def dumps(report: list[Record], sorted_groups: bool = True) -> str:
    def resolve(states):
        if S.FAILED in states:
            status = S.FAILED
        elif S.WARN in states:
            status = S.WARN
        elif S.OK in states:
            status = S.OK
        else:
            status = S.NOSTATUS
        return status

    def color(status: S, fall=".") -> ColorText:
        message = {
            S.OK: "+",
            S.FAILED: "x",
            S.WARN: "!",
            S.NOSTATUS: ".",
        }[status]
        return ColorText(message, status)

    def colorize(message: str, status: S) -> ColorText:
        return ColorText(message, status)

    pre = " " * 3
    result = []

    width = max(len(record.key) for record in report)

    def bygroup(fn, items):
        return fn(items, key=lambda r: r.group)

    def nosorted(items, key):
        return items

    for group, itrecords in bygroup(itertools.groupby, bygroup(nosorted, report)):
        records = list(itrecords)
        status = resolve(set(record.status for record in records))
        result.append(f"{color(status)} {group}")
        start = " " * 4 + " " * width
        for record in records:
            if record.group != group:
                continue
            if isinstance(record.report, str):
                message = record.report
            elif isinstance(record.report, list):
                message = indent("\n".join(record.report), pre=start).lstrip()
            else:
                raise RuntimeError("unable to handle type", type(record.report))
            result.append(f"{pre}{colorize(record.key, record.status):{width}} {message}")

    return "\n".join(result)


def print_report(report: list[Record], sorted_groups: bool = True) -> int:
    errors = sum(r.status == S.FAILED for r in report)
    warnings = sum(r.status == S.WARN for r in report)
    print(dumps(report, sorted_groups))
    if errors:
        t = ColorText("FAILED", S.FAILED)
        print(f"{t} found {errors} errors, and {warnings} warnings")
    elif warnings:
        t = ColorText("WARN", S.WARN)
        print(f"{t} found {warnings} warnings")
    else:
        t = ColorText("OK", S.OK)
        print(f"{t}")
    return min(int(sum(r.status == S.FAILED for r in report)), 1)


def check(fn):
    @functools.wraps(fn)
    def _fn(*args, **kwargs):
        result = fn(*args, **kwargs)
        return result if isinstance(result, list) else [result]

    return _fn


if __name__ == "__main__":
    report = []
    report.append(Record(S.NOSTATUS, "group-0", "key-0", "message"))
    report.append(Record(S.OK, "group-0", "key-2", "message-2"))
    report.append(Record(S.FAILED, "group-0", "key-3", "message-3"))
    report.append(Record(S.WARN, "group-0", "key-4", "message-4"))
    report.append(Record(S.NOSTATUS, "group-1", "key-0", "message"))
    report.append(Record(S.OK, "group-1", "key-2", "message-2"))
    report.append(Record(S.OK, "group-1", "key-3", "message-3"))
    report.append(Record(S.WARN, "group-1", "key-4", "message-4"))
    ret = print_report(report)
    print(f"Final status -> {ret}")



from __future__ import annotations

import dataclasses as dc
import logging
import os
import subprocess
import sys
import threading
import time
from pathlib import Path
from typing import BinaryIO, Literal, Sequence

COLORS = {
    "blue": "\033[94m",
    "green": "\033[92m",
    "red": "\033[91m",
    "clear": "\033[0m",
}

logger = logging.getLogger(__name__)


class RunnerError(Exception):
    pass


@dc.dataclass
class BaseFilter:
    def __call__(self, stream: BinaryIO) -> None:
        for line in iter(stream.readline, b""):
            pass


OMode = Literal["capture", "null", "display", "capture+display"] | BaseFilter
EMode = Literal["null", "display"] | BaseFilter
Paths = str | Path | Sequence[str | Path]


@dc.dataclass
class CaptureFilter(BaseFilter):
    encode: str | None = "utf-8"
    result: str | bytes | None = None

    def __call__(self, stream: BinaryIO) -> None:
        result = []
        for line in iter(stream.readline, b""):
            result.append(line[:-1])
        stream.close()
        if self.encode:
            self.result = b"\n".join(result).decode(self.encode)
        else:
            self.result = b"\n".join(result)


@dc.dataclass
class DisplayFilter(BaseFilter):
    color: str | None
    pre: str = "   | "
    clear: str = COLORS["clear"]
    capture: bool = False
    encode: str | None = "utf-8"
    result: str | bytes | None = None

    def __call__(self, stream: BinaryIO) -> None:
        result = []
        for rawline in iter(stream.readline, b""):
            if self.capture:
                result.append(rawline[:-1])
            line = rawline.decode("utf-8")
            if line.strip().startswith("Warning:"):
                line = line.replace("Warning:", f"{COLORS['red']}Warning:{self.clear}{self.color}")
            print(
                f"{self.pre}{self.color}{line.rstrip()}{self.clear}",
                flush=True,
                file=sys.stderr,
            )
        stream.close()
        if self.capture:
            self.result = b"\n".join(result).decode(self.encode) if self.encode else b"\n".join(result)


def mkpaths(args: Paths) -> list[str]:
    return [str(args)] if isinstance(args, (str, Path)) else [str(a) for a in args]


def runc(
    args: Paths,
    stdout: OMode = "display",
    stderr: EMode = "display",
    overrides: dict[str, str] | None = None,
    **kwargs,
) -> str | bytes | None:
    kwargs["env"] = kwargs.pop("env") if "env" in kwargs else os.environ.copy()
    kwargs["env"].update(overrides or {})
    kwargs["cwd"] = (str(v) if (v := kwargs.pop("cwd")) else None) if "cwd" in kwargs else None

    with subprocess.Popen(mkpaths(args), stderr=subprocess.PIPE, stdout=subprocess.PIPE, **kwargs) as process:
        if stdout == "capture":
            ofiltermap: BaseFilter = CaptureFilter()
        elif stdout == "null":
            ofiltermap = BaseFilter()
        elif stdout == "display":
            ofiltermap = DisplayFilter(COLORS["blue"], "   | ")
        elif stdout == "capture+display":
            ofiltermap = DisplayFilter(COLORS["blue"], "   | ", capture=True)
        elif isinstance(stdout, BaseFilter):
            ofiltermap = stdout
        else:
            raise RuntimeError(f"unsupported type in {stdout=}")
        othread = threading.Thread(target=ofiltermap, args=(process.stdout,), daemon=True)

        if stderr == "null":
            efiltermap: BaseFilter = BaseFilter()
        elif stderr == "display":
            efiltermap = DisplayFilter(COLORS["green"], "   | ")
        elif isinstance(stderr, BaseFilter):
            efiltermap = stderr
        else:
            raise RuntimeError(f"unsupported type in {stderr=}")
        ethread = threading.Thread(
            target=efiltermap,
            args=(process.stderr,),
            daemon=True,
        )
        othread.start()
        ethread.start()
        while process.poll() is not None:
            time.sleep(0.05)
        othread.join()
        ethread.join()

    if process.returncode:
        envs = " ".join(f'{k}="{v}"' for k, v in (overrides or {}).items())
        cmdline = subprocess.list2cmdline(mkpaths(args))
        raise RunnerError(f"failed to execute in cwd={kwargs['cwd']} ==> {envs} {cmdline}")
    return ofiltermap.result if hasattr(ofiltermap, "result") else None


if __name__ == "__main__":
    x = runc(["ls", "-l"], "capture", "display")
    print(x)



import functools
import logging
from argparse import Namespace
from typing import Callable

import click

TypeFn = Callable[[Namespace], None]


def add_loglevel(fn: TypeFn) -> TypeFn:
    fn = click.option("-v", "--verbose", count=True)(fn)
    fn = click.option("-q", "--quiet", count=True)(fn)
    return fn


def process_loglevel(options: Namespace, verbose_flag: bool = False) -> Namespace:
    level = max(min(options.__dict__.pop("verbose") - options.__dict__.pop("quiet"), 1), -1)

    # console = Console(theme=Theme({"log.time": "cyan"}))
    logging.basicConfig(
        level={-1: logging.WARNING, 0: logging.INFO, 1: logging.DEBUG}[level],
        # datefmt="[%X]",
        # handlers=[RichHandler(console=console, rich_tracebacks=True)]
    )
    if verbose_flag:
        options.verbose = level > 0
    return options


def clickwrapper(
    add_arguments: Callable[[TypeFn], TypeFn] | None = None,
    process_options: Callable[[Namespace], None | Namespace] | None = None,
    verbose_flag: bool = False,
) -> Callable[[TypeFn], None]:
    def _clickwrapper(fn: TypeFn):
        fn = add_loglevel(fn)
        if add_arguments:
            fn = add_arguments(fn)

        @functools.wraps(fn)
        def __clickwrapper(*args, **kwargs):
            options = Namespace(**kwargs)
            options = process_loglevel(options, verbose_flag=verbose_flag) or options
            if hasattr(options, "error"):
                raise RuntimeError("you have an error option")

            def error(msg):
                raise click.UsageError(msg)

            options.error = error
            if process_options:
                options = process_options(options) or options

            return fn(options)

        return __clickwrapper

    return _clickwrapper



from __future__ import annotations

import dataclasses as dc
from pathlib import Path

from .runner import Paths, Runner


@dc.dataclass
class Git:
    workdir: Path
    runc: Runner

    @classmethod
    def new(cls, workdir: Path, verbose: bool) -> Git:
        workdir = workdir.expanduser().absolute()
        return cls(workdir, runc=Runner(verbose=verbose, exe=["git", "--git-dir", f"{workdir}/.git"]))

    # @classmethod
    # def clone(cls, url: str, workdir: Path, args: Paths | None = None, verbose: bool = False) -> Git:
    #     runc = Runner(verbose=True)
    #     runc(["git", "clone", *(args or []), url, workdir])
    #     return cls.new(workdir, verbose)

    def __call__(self, args: Paths) -> str:
        out = self.runc(args, capture=True) or ""
        return (out.decode("utf-8") if isinstance(out, bytes) else out).strip()

    def branch(self, name: str = "HEAD") -> str:
        return self(["rev-parse", "--abbrev-ref", name])

    def commits_on_branch(self, main: str = "origin/main"):
        changes = self(["reflog", "show", "--no-abbrev", self.branch()])
        return len([n for n in changes.split("\n") if n.strip()])
        # parent = self(["merge-base", self.branch(), main]).strip()
        # return int(self(["rev-list", "--count", f"{parent}..{self.branch()}"]))



import argparse

import pytest

from acbox.cli import shared
from acbox.cli.flags.base import ArgumentTypeBase
from acbox.cli.parsers.base import ArgumentParserBase


class Flag(ArgumentTypeBase):
    def __init__(self, only_odd=False):
        self.only_odd = only_odd
        super().__init__()

    def validate(self, txt):
        value = int(txt)
        if self.only_odd:
            if (value % 2) == 0:
                raise argparse.ArgumentTypeError(f"non odd value '{txt}'")
        return value


def test_check_default_constructor():
    """verifies the class constructor requires no arguments"""

    class A:
        def __init__(self, value):
            pass

    pytest.raises(RuntimeError, shared.check_default_constructor, A)

    class A:
        def __init__(self, value=1):
            pass

    assert shared.check_default_constructor(A) is None


def test_add_argument():
    """test all failures mode for add_argument default"""
    p = ArgumentParserBase([], exit_on_error=False)

    with pytest.raises(ValueError) as e:
        p.add_argument("--flag", type=Flag, default="yyy")
    assert e.value.args[0] == "invalid literal for int() with base 10: 'yyy'"

    with pytest.raises(RuntimeError) as e:
        p.add_argument("--flag", type=Flag(only_odd=True), default="124")
    assert e.value.args[0] == "cannot use value='124' as default: non odd value '124'"

    with pytest.raises(RuntimeError) as e:
        p.add_argument("--flag", type=Flag(only_odd=True), default=124)
    assert e.value.args[0] == "cannot use value=124 as default: non odd value '124'"

    with pytest.raises(RuntimeError) as e:
        p.add_argument("--flag", type=Flag(only_odd=True), default=126)
    assert e.value.args[0] == "cannot use value=126 as default: non odd value '126'"

    with pytest.raises(RuntimeError) as e:
        p.add_argument("--flag", type=Flag(only_odd=True), default="126")
    assert e.value.args[0] == "cannot use value='126' as default: non odd value '126'"

    p.add_argument("--flag", type=Flag)
    assert len(action := [a for a in p._actions if isinstance(a.type, ArgumentTypeBase)]) == 1
    assert action[0].default.default is ArgumentTypeBase._NA

    p.add_argument("--flag1", type=Flag, default="124")
    assert len(actions := [a for a in p._actions if isinstance(a.type, ArgumentTypeBase)]) == 2
    assert actions[1].default.default == 124

    p.add_argument("--flag2", type=Flag, default=126)
    assert len(actions := [a for a in p._actions if isinstance(a.type, ArgumentTypeBase)]) == 3
    assert actions[2].default.default == 126


def test_special_flag_no_restriction():
    """parse arguments with no default and no constrain on Flag"""
    p = ArgumentParserBase([], exit_on_error=False)
    p.add_argument("--flag", type=Flag)

    a = p.parse_args([])
    assert a.flag is None

    a = p.parse_args(
        [
            "--flag",
            "123",
        ]
    )
    assert a.flag == 123

    a = p.parse_args(
        [
            "--flag",
            "124",
        ]
    )
    assert a.flag == 124

    with pytest.raises(argparse.ArgumentError) as e:
        p.parse_args(["--flag", "boo"])
    assert e.value.args[-1] == "invalid Flag value: 'boo'"

    # same as above but with a default fallback
    p = ArgumentParserBase([], exit_on_error=False)
    p.add_argument("--flag", type=Flag, default="42")

    a = p.parse_args([])
    assert a.flag == 42

    a = p.parse_args(
        [
            "--flag",
            "123",
        ]
    )
    assert a.flag == 123

    a = p.parse_args(
        [
            "--flag",
            "124",
        ]
    )
    assert a.flag == 124

    with pytest.raises(argparse.ArgumentError) as e:
        p.parse_args(["--flag", "boo"])
    assert e.value.args[-1] == "invalid Flag value: 'boo'"


def test_special_flag_with_restriction_no_default():
    p = ArgumentParserBase([], exit_on_error=False)
    p.add_argument("--flag", type=Flag(only_odd=True))

    a = p.parse_args([])
    assert a.flag is None

    a = p.parse_args(
        [
            "--flag",
            "123",
        ]
    )
    assert a.flag == 123

    with pytest.raises(argparse.ArgumentError) as e:
        p.parse_args(["--flag", "122"])
    assert e.value.args[-1] == "non odd value '122'"

    with pytest.raises(argparse.ArgumentError) as e:
        p.parse_args(["--flag", "boo"])
    assert e.value.args[-1] == "invalid Flag value: 'boo'"

    # same as above but with a default fallback
    p = ArgumentParserBase([], exit_on_error=False)
    p.add_argument("--flag", type=Flag(only_odd=True), default=43)

    a = p.parse_args([])
    assert a.flag == 43

    a = p.parse_args(
        [
            "--flag",
            "123",
        ]
    )
    assert a.flag == 123

    with pytest.raises(argparse.ArgumentError) as e:
        p.parse_args(["--flag", "122"])
    assert e.value.args[-1] == "non odd value '122'"

    with pytest.raises(argparse.ArgumentError) as e:
        p.parse_args(["--flag", "boo"])
    assert e.value.args[-1] == "invalid Flag value: 'boo'"



from __future__ import annotations

import sys

from acbox import runner


def test_stderr_stdout(resolver):
    exe = resolver.lookup("test-script.py")

    runc = runner.Runner(verbose=False)
    out = runc([sys.executable, exe], capture=True)

    assert (
        out.strip().replace("\r", "")
        == """
HELLO=N/A
line (out) 1
line (out) 2
line (out) 4
line (out) 5
line (out) 7
line (out) 8
""".strip().replace("\r", "")
    )

    out = runc([sys.executable, exe], overrides={"HELLO": "123"}, capture=True)
    assert (
        out.strip().replace("\r", "")
        == """
HELLO=123
line (out) 1
line (out) 2
line (out) 4
line (out) 5
line (out) 7
line (out) 8
""".strip().replace("\r", "")
    )



def test_script_lookup(resolver):
    path = resolver.lookup("simple-script.py")
    assert path
    assert path.exists()


def test_script_load(resolver):
    data = resolver.load("simple-script.py", "text")
    assert (
        data
        == """
import sys

VALUE = 123


def hello(msg):
    print(f"Hi {msg}")


def hello_stderr(msg):
    print(f"Hi stderr {msg}", file=sys.stderr)


if __name__ == "__main__":
    hello("Antonio")
    hello_stderr("Antonio")
""".lstrip()
    )

    mod = resolver.load("simple-script.py", "mod")
    assert mod.VALUE, 123



from __future__ import annotations

import argparse
import types

from ..flags.base import ArgumentTypeBase
from ..shared import ArgsCallback, check_default_constructor


class ArgumentParserBase(argparse.ArgumentParser):
    def __init__(self, modules: list[types.ModuleType], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.modules = modules
        self.callbacks: list[ArgsCallback | None] = []

    def _post_process(self, options):
        for name in dir(options):
            if isinstance(getattr(options, name), ArgumentTypeBase):
                fallback = getattr(options, name).value
                setattr(
                    options,
                    name,
                    None if fallback is ArgumentTypeBase._NA else fallback,
                )
        for callback in self.callbacks:
            options = callback(options) or options
        return options

    def parse_known_args(self, args=None, namespace=None):
        options, argv = super().parse_known_args(args, namespace)
        return self._post_process(options), argv

    def parse_args(self, args=None, namespace=None):
        options = super().parse_args(args, namespace)
        for name in dir(options):
            if isinstance(getattr(options, name), ArgumentTypeBase):
                fallback = getattr(options, name).value
                setattr(
                    options,
                    name,
                    None if fallback is ArgumentTypeBase._NA else fallback,
                )
        return options

    def add_argument(self, *args, **kwargs):
        typ = kwargs.get("type")
        obj = None
        if isinstance(typ, type) and issubclass(typ, ArgumentTypeBase):
            check_default_constructor(typ)
            obj = typ()
        if isinstance(typ, ArgumentTypeBase):
            obj = typ
        if obj is not None:
            obj.default = kwargs.get("default", ArgumentTypeBase._NA)
            kwargs["default"] = obj
            kwargs["type"] = obj
        super().add_argument(*args, **kwargs)

    def error(self, message):
        try:
            super().error(message)
        except SystemExit:
            # gh-121018
            raise argparse.ArgumentError(None, message)

    @classmethod
    def get_parser(cls, modules: list[types.ModuleType], **kwargs):
        raise NotImplementedError("implement this method")
